import json

from traceback import format_exc
from functools import partial
from sqlor.dbpools import DBPools
from appPublic.streamhttpclient import StreamHttpClient, liner
from appPublic.log import debug, exception, error
from ahserver.globalEnv import password_decode

def get_dbname():
	f = get_serverenv('get_module_dbname')
	dbname = f('uapi')
	return dbname

class UAPI(StreamHttpClient):
	def __init__(self, appid, env={}):
		self.appid = appid
		super().__init__()
		self.te = MyTemplateEngine([], env=env)
		self.env = env
		self.auth_api = None
		self.auth_ret = None
	
	async def tmplrender(self, tmplstr, params={}):
		if tmplstr is None:
			return None
		ns = self.env.copy()
		ns.update(params)
		te = MyTemplateEngine([], env=self.env)
		return te.renders(headerstmpl, ns)

	async def make_call(self, request, upapiid, callerid, 
							params={}, 
							stream=False, 
							env={}):
		self.env.update(env)
		db = DBPools()
		dbname = get_dbname()
		uapi = None
		auth_uapi = None
		async with db.sqlorContext(dbname) as sor:
			upapi = await self.get_upappapis(upapiid)
			kinfo = await self.get_orgapikey(sor, upapi.upappid, callerid)
			self.env.update(kinfo)
			auth_uapi = None
			if kinfo.auth_apiid:
				auth_upapi = await self.get_upappapis(kinfo.auth_apiid)
				auth_uapi = await self.get_uapi(sor, kinfo.auth_upapi.uapiid)
			uapi = await self.get_uapi(sor, upapi.uapiid)
		if auth_uapi:
			await self.do_auth(auth_uapi)
		if stream:
			return await self.do_stream_call(self, request, uapi, params=params) 
		else:
			return await self.do_call(self, request, uapi, params=params) 

	async def do_auth(self, auth_uapi):
		b = await self.do_call(auth_uapi)
		d = json.loads(b.encode('utf-8'))
		self.env.update(d)
		return
	
	async def do_call(self, request, api, params={}):
		url = self.env.get('baseurl') + api.path
		method = auth_uapi.httpmethod
		header = self.rendertmpl(api.headers, params=params)
		header = json.loads(headers)
		body = self.rendertmpl(api.body, params=params)
		_params = self.rendertmpl(api.params, params=params)
		if _params:
			_params = json.loads(_params)
		b = await self.request(method, url, 
							headers=headers,
							body=body,
							params=_params)

		d = json.loads(b.encode('utf-8'))
		if api.response:
			return self.rendertmpl(api.response, params=d)
		return d

	async def do_stream_call(self, request, api, params={}):
		self.resp = web.StreamResponse()
		await self.resp.prepare(request)
		url = self.env.get('baseurl') + api.path
		method = auth_uapi.httpmethod
		header = self.rendertmpl(api.headers, params=params)
		header = json.loads(headers)
		body = self.rendertmpl(api.body, params=params)
		if body:
			bdy = json.loads(body)
			bdy['stream'] = True
			body = bdy.dumps(bdy, ensure_ascii=False)
		_params = self.rendertmpl(api.params, params=params)
		if _params:
			_params = json.loads(_params)
		hc = StreamHttpClient()
		gen = hc(method, url, 
						headers=headers,
						body=body,
						params=_params)
		self.chunk_match = self.api.chunk_match or ''
		async for line in liner(gen):
			self.streamline_handle(line)
		return self.resp

	async def streamline_handle(self, line, resptmpl):
		cnt = len(self.chunk_match)
		line = line[cnt:]
		try:
			dic = json.loads(line)
			if resptmpl:
				jstr = self.rendertmpl(resptmpl, dic) + '\n'
			else:
				jstr = json.dumps(self.resp_json, ensure_ascii=False) + '\n'
			bin = jstr.encode('utf-8')
			await self.resp.write(bin)
			await self.resp.drain()
		except Exception as e:
			exception(f'{line=}\n{e=},{format_exc()}')
			return
			
	async def get_upappapis(self, sor, upapiid):
		recs = await sor.R('upappapis', {'id': upapiid})
		if len(recs) < 1:
			e = Exception(f'{upapiid} not found in table(upappapis)')
			exception(f'{e}\n{format_exc()}')
			raise e
		return recs[0]
	
	async def get_uapi(self, sor, uapiid):
		recs = await sor.R('uapi', {'id': uapiid})
		if len(recs) < 1:
			e = Exception(f'{uapiid} not found in table(uapi)')
			exception(f'{e}\n{format_exc()}')
			raise e
		return recs[0]
		
	async def get_orgapikey(self, sor, upappid, callerid):
		"""
		argumemts:
			upappid: upappid which will make call to
			orgid: owner organization or user which as the caller of the call
		return:
			None: this orgid has not gotton apikey from upapp
			dict if apikey and upapp infos
		"""
		sql = """select 
a.myappid, 
a.apisetid, 
a.auth_apiid,
a.baseurl,
b.*,
from upapp a, upappkey b 
where a.id = b.upappid
	and a.id = ${appid}$
	and b.ownerid = ${orgid}$"""
		recs = await sor.sqlExe(sql, {'appid':upappid, 'orgid': callerid})
		if len(recs) < 1:
			e = Exception(f'{appid=}, {callerid=} has not apikey')
			exception(f'{e}, {format_exc()}')
			raise e
		r = recs[0]
		return {
			'apikey':r.apikey,
			'secretkey':r.secretkey,
			'apisetid': r.apisetid,
			'auth_apiid': r.auth.apiid,
			'myappid': r.myappid
		}
	
	async def get_authapi(self, sor):
		sql = "select * from jsonhttpapi where name = 'auth' and appid=${appid}$"
		recs = await sor.sqlExe(sql, {'appid': self.appid})
		if len(recs):
			return recs[0]
		return None

	async def get_apiinfo(self, orgids, apiname):
		db = DBPools()
		dbname = get_serverenv('get_module_dbname')('uapi')
		async with db.sqlorContext(dbname) as sor:
			sql = """select c.baseurl,
c.name as appname,
b.*,
a.apikey,
a.secretkey
from upappkey a, jsonhttpapi b, upapp c
where
	a.upappid=b.appid
	and a.upappid = c.id
	and b.appid = ${appid}$
	and b.name = ${apiname}$
	and a.ownerid = ${orgid}$
"""
			for orgid in orgids:
				ns = {
					"orgid": orgid,
					"apiname": appname,
					"appid": self.appid
				}
				recs = await sor.sqlExe(sql, ns)
				if len(recs) > 0:
					rec = recs[0]
					rec.apikey = password_decode(rec.apikey)
					if rec.secretkey:
						rec.secretkey = password_decode(rec.secretkey)
					if rec.need_auth and self.auth_api is None:
						self.auth_api = await self.get_authapi(sor)
					if rec.need_auth and self.auth_api is None:
						e = Exception(f'{rec.appname} app has not auth api but need')
						exception(f'{e=}, {format_exc()}')
						raise e
					rec.auth_api = self.auth_api
					rec.apikey_orgid = orgid
					return recs[0]
		return None

